<?php namespace RVR\Controller;

use DateTime;
use RVR\PersistentData\Model\OAuthSession;
use RVR\PersistentData\Model\User;
use RVR\Repository\OAuthClientRepository;
use SokoWeb\Interfaces\Authentication\IAuthenticationRequired;
use SokoWeb\Interfaces\Response\IRedirect;
use SokoWeb\Response\Redirect;
use SokoWeb\Response\HtmlContent;

class OAuthSessionController implements IAuthenticationRequired
{
    private OAuthClientRepository $oAuthClientRepository;

    public function __construct()
    {
        $this->oAuthClientRepository = new OAuthClientRepository();
    }

    public function isAuthenticationRequired(): bool
    {
        return true;
    }

    public function auth(): HtmlContent|Redirect
    {
        $redirectUri = \Container::$request->query('redirect_uri');
        $clientId = \Container::$request->query('client_id');
        $scope = \Container::$request->query('scope') ? \Container::$request->query('scope'): '';
        $state = \Container::$request->query('state');
        $nonce = \Container::$request->query('nonce') ? \Container::$request->query('nonce'): '';
        $codeChallenge = \Container::$request->query('code_challenge') ?: null;
        $codeChallengeMethod = \Container::$request->query('code_challenge_method') ?: null;

        if (!$clientId || !$redirectUri || !$state || (!$codeChallenge && $codeChallengeMethod)) {
            return new HtmlContent('oauth/oauth_error', ['error' => 'An invalid request was made. Please start authentication again.']);
        }

        if ($codeChallenge && (strlen($codeChallenge) < 43 || strlen($codeChallenge) > 128)) {
            return new HtmlContent('oauth/oauth_error', ['error' => 'Code challenge should be one between 43 and 128 characters long.']);
        }
        $possibleCodeChallengeMethods = ['plain', 'S256'];
        if ($codeChallenge && !in_array($codeChallengeMethod, $possibleCodeChallengeMethods)) {
            return new HtmlContent('oauth/oauth_error', ['error' => 'Code challenge method should be one of the following: ' . implode(',', $possibleCodeChallengeMethods)]);
        }

        $client = $this->oAuthClientRepository->getByClientId($clientId);
        if ($client === null) {
            return new HtmlContent('oauth/oauth_error', ['error' => 'Client is not found.']);
        }

        $redirectUriQueryParsed = [];
        if (str_contains('?', $redirectUri)) {
            [$redirectUriBase, $redirectUriQuery] = explode('?', $redirectUri, 2);
            parse_str($redirectUriQuery, $redirectUriQueryParsed);
        } else {
            $redirectUriBase = $redirectUri;
        }
        if (!in_array($redirectUriBase, $client->getRedirectUrisArray())) {
            return new HtmlContent('oauth/oauth_error', ['error' => 'Redirect URI \'' . $redirectUriBase .'\' is not allowed for this client.']);
        }

        /**
        * @var ?User $user
        */
        $user = \Container::$request->user();
        $code = bin2hex(random_bytes(16));

        $session = new OAuthSession();
        $session->setClientId($clientId);
        $session->setNonce($nonce);
        $session->setScope($scope);
        $session->setCodeChallenge($codeChallenge);
        $session->setCodeChallengeMethod($codeChallengeMethod);
        $session->setUser($user);
        $session->setCode($code);
        $session->setCreatedDate(new DateTime());
        $session->setExpiresDate(new DateTime('+5 minutes'));
        \Container::$persistentDataManager->saveToDb($session);

        $redirectUriQueryParsed = array_merge($redirectUriQueryParsed, [
            'state' => $state,
            'code' => $code
        ]);
        $finalRedirectUri = $redirectUriBase . '?' . http_build_query($redirectUriQueryParsed);

        return new Redirect($finalRedirectUri, IRedirect::TEMPORARY);
    }
}
